/*
 * yspl.i --
 *
 * Yorick interface for Signal Processing Library (SPL).
 *
 *-----------------------------------------------------------------------------
 *
 * Copyright (C) 2009-2010 Eric Thiébaut <thiebaut@obs.univ-lyon1.fr>
 *
 * This software is governed by the CeCILL-C license under French law and
 * abiding by the rules of distribution of free software.  You can use, modify
 * and/or redistribute the software under the terms of the CeCILL-C license as
 * circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty and the software's author, the holder of the
 * economic rights, and the successive licensors have only limited liability.
 *
 * In this respect, the user's attention is drawn to the risks associated with
 * loading, using, modifying and/or developing or reproducing the software by
 * the user in light of its specific status of free software, that may mean
 * that it is complicated to manipulate, and that also therefore means that it
 * is reserved for developers and experienced professionals having in-depth
 * computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling
 * the security of their systems and/or data to be ensured and, more
 * generally, to use and operate it in the same conditions as regards
 * security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-C license and that you accept its terms.
 *
 *-----------------------------------------------------------------------------
 *
 * $Id$
 * $Log$
 *-----------------------------------------------------------------------------
 */

if (is_func(plug_in)) plug_in, "yspl";

//extern spl_get_version;
/* DOCUMENT spl_get_version();
 *
 *   Returns the version string of the Yorick interface to the Signal
 *   Processing Library (SPL).
 */

//extern spl_get_info_level;
//extern spl_get_info_text;
/* DOCUMENT spl_get_info_level(info);
 *     -or- spl_get_info_text(info);
 */

local spl_blas;
/* DOCUMENT spl_blas -- Basic Linear Algebra Subroutines

     spl_amax(x)       - get maximum absolute value of X.
     spl_asum(x)       - get the sum of the absolute values of X.
     spl_axpy(a, x, y) - compute A.X + Y.
     spl_copy, x, y;   - copy a vector into another one.
     spl_dot(x, y)     - get the dot product of X and Y.
     spl_iamax(x)      - get index of maximum absolute value of X.
     spl_nrm2(x)       - get the Euclidean norm of X.
     spl_scal, a, x;   - scale X by scalar A.
     spl_sum(x)        - get the sum of the values of X.
     spl_swap, x, y;   - exchange the contents of X and Y.
     spl_zero, x;      - fill an array with zeroes.

   SEE ALSO: spl_amax, spl_asum, spl_axpy, spl_copy, spl_dot, spl_iamax,
             spl_nrm2, spl_scal, spl_sum, spl_swap, spl_zero. */

extern spl_asum;
/* DOCUMENT spl_asum(x);
     Returns the sum of the absolute values of X, that is: sum(abs(X)).
     This is also the L1-norm of X.

   SEE ALSO: spl_blas. */

extern spl_sum;
/* DOCUMENT spl_sum(x);
     Returns the sum of the values of X, that is: sum(X).

   SEE ALSO: spl_blas. */

extern spl_amax;
/* DOCUMENT spl_amax(x);
     Returns maximum absolute value of X, that is: max(abs(X)).
     This is also the infinite-norm of X.

   SEE ALSO: spl_blas. */

local spl_enorm;
extern spl_nrm2;
/* DOCUMENT spl_nrm2(x);
       -or- spl_enorm(x);
     Returns the Euclidean (L2) norm of X, that is: sqrt(sum(X*X)) but
     avoiding overflows.  This is also the L2-norm of X.

   SEE ALSO: spl_blas. */
spl_enorm = spl_nrm2;

extern spl_dot;
/* DOCUMENT spl_dot(x, y);
       -or- spl_dot(x, y, i);
       -or- spl_dot(w, x, y);
       -or- spl_dot(f, x, y);
       -or- spl_dot(n, x, offx, incx, y, offy, incy);
     Returns the dot product of "vectors" X and Y, that is sum(X*Y).

   SEE ALSO: spl_blas. */

//extern spl_axpy;
/* DOCUMENT spl_axpy, a, x, y;
       -or- spl_axpy(a, x, y);

   SEE ALSO: spl_blas. */

//extern spl_copy;
/* DOCUMENT spl_copy, x, y;
     Copy X into Y.

   SEE ALSO: spl_blas. */

//extern spl_swap;
/* DOCUMENT spl_swap, x, y;
     Exchange contents of X and Y.

   SEE ALSO: spl_blas. */

//extern spl_zero;
/* DOCUMENT spl_zero, x;
     Fill X with zeroes.

   SEE ALSO: spl_blas. */

//extern spl_scal;
/* DOCUMENT spl_scal, a, x;
       -or- spl_scal(a, x);
     Scale the values of X by scalar A.  When called as a subroutine, the
     operation is done in-place; otherwise, the function returns a new
     array with the result.

   SEE ALSO: spl_blas. */

/*---[ BLAS constants ]------------------------------------------------------*/

local SPL_BLAS_ROW_MAJOR, SPL_BLAS_COL_MAJOR;
local SPL_BLAS_NO_TRANS, SPL_BLAS_TRANS, SPL_BLAS_CONJ_TRANS;
local SPL_BLAS_UPPER, SPL_BLAS_LOWER;
local SPL_BLAS_NON_UNIT, SPL_BLAS_UNIT;
local SPL_BLAS_LEFT, SPL_BLAS_RIGHT;
/* DOCUMENT SPL_BLAS_ROW_MAJOR
       -or- SPL_BLAS_COL_MAJOR
     Storage order for packed matrices.

            SPL_BLAS_NO_TRANS
       -or- SPL_BLAS_TRANS
       -or- SPL_BLAS_CONJ_TRANS
     Transpose flag.

            SPL_BLAS_UPPER
       -or- SPL_BLAS_LOWER
     Part to consider for triangular matrices.

            SPL_BLAS_NON_UNIT
       -or- SPL_BLAS_UNIT
     Diagonal flag.

            SPL_BLAS_LEFT
       -or- SPL_BLAS_RIGHT
     Side flag.

   SEE ALSO:
 */

/* The following values follow the standard definitions. */
SPL_BLAS_ROW_MAJOR  = 101n;
SPL_BLAS_COL_MAJOR  = 102n;
SPL_BLAS_NO_TRANS   = 111n;
SPL_BLAS_TRANS      = 112n;
SPL_BLAS_CONJ_TRANS = 113n;
SPL_BLAS_UPPER      = 121n;
SPL_BLAS_LOWER      = 122n;
SPL_BLAS_NON_UNIT   = 131n;
SPL_BLAS_UNIT       = 132n;
SPL_BLAS_LEFT       = 141n;
SPL_BLAS_RIGHT      = 142n;

/*---------------------------------------------------------------------------*/
/* CONJUGATE GRADIENT */

local SPL_CG_ERROR, SPL_CG_START, SPL_CG_RESTART, SPL_CG_NEWX;
local SPL_CG_PRECOND, SPL_CG_AP, SPL_CG_FINISH;
local SPL_CG_NON_CONVEX, SPL_CG_TRUNCATED; // FIXME:
/*
 * DOCUMENT    The possible values of Conjugate Gradient state.
 *
 *   SPL_CG_ERROR      - error;
 *   SPL_CG_START      - start with no initial variable (all zero);
 *   SPL_CG_RESTART    - start or restart with initial variables;
 *   SPL_CG_NEWX       - new step taken;
 *   SPL_CG_PRECOND    - caller has to to compute z = Q.r;
 *   SPL_CG_AP         - caller has to to compute q = A.p;
 *   SPL_CG_FINISH     - convergence or rounding errors prevent further
 *                       progress;
 *   SPL_CG_NON_CONVEX - non-positive definitiveness has been detected;
 *   SPL_CG_TRUNCATED  - step truncated at trust region boundary.
 */
SPL_CG_ERROR      = -1n;
SPL_CG_START      =  0n;
SPL_CG_RESTART    =  1n;
SPL_CG_NEWX       =  2n;
SPL_CG_PRECOND    =  3n;
SPL_CG_AP         =  4n;
SPL_CG_FINISH     =  5n;
SPL_CG_NON_CONVEX =  6n;
SPL_CG_TRUNCATED  =  7n;

extern spl_cgstp;
/* DOCUMENT spl_cgstp, n, p, q, r, x, rho, state;
         or spl_cgstp, n, p, q, r, x, z, rho, state;

      The routine spl_cgstp performs a step of the (preconditioned) linear
      conjugate gradient method to solve the linear system A.x = b.

      N is the size of the problem.

      P is a floating point work array with N elements; see parameter STATE
        for explanations.

      Q is a floating point work array of size N; on return with STATE set to
        SPL_CG_AP, the caller must compute the matrix vector multiplication
        q = A.p (i.e. store into Q the result of multiplying P by the matrix
        A); on return with STATE set to SPL_CG_NEWX, Q is filled with the
        unscaled residuals step.

      R is a floating point array with N elements; on initialization or
        restart (with state set to SPL_CG_START or SPL_CG_RESTART), R stores
        the right hand side vector b; otherwise, R stores the current
        residuals.

      X is a floating point array with N elements to store the unknowns.  On
        initialization with given parameter values or restart, i.e. with STATE
        set to SPL_CG_RESTART, X stores the initial unknowns x0; on initial
        entry with STATE set to SPL_CG_START, the contents of X is ignored (X
        will be filled with zeros); otherwise, X stores the current
        solution.

      Z is an optional floating point work array with N elements; on return
        with state SPL_CG_PRECOND, the caller must compute and store the
        preconditioned residuals into Z (given the current residuals which are
        in R).  If Z is specified, the preconditioned conjugate gradient
        method is used; otherwise, the unpreconditioned conjugate gradient
        method is used and SPL_CG_PRECOND will never get returned.  Z can be
        the same array as Q to save memory.

      RHO is floating point array of 4 elements used to store the dot product
      r'.z of the current and previous iterations and to store some parameters
      of the algorithm.  RHO(1) is the former value of RHO(2); RHO(2) is the
      dot product of the residuals R by the preconditioned residuals Z or by R
      if no preconditioner is used; RHO(3) is the optimal step size (ALPHA);
      RHO(4) is the weight of the former search direction (BETA).

      STATE is a variable with integer scalar value.  On initialization or
      restart of the conjugate gradient iterations, STATE must be set to
      SPL_CG_START or SPL_CG_RESTART and R must be set to the right hand side
      vector b; if STATE is SPL_CG_START, X is filled with zeros by the
      routine; if STATE is SPL_CG_RESTART, X must be set with some initial
      values of the parameters.  Upon return of the function, the value stored
      in STATE is one of:

      SPL_CG_NEWX: X stores the current solution, R stores the current
         residuals, the change in parameters is equal to RHO(3)*P and the
         corresponding residuals change is equal to -RHO(3)*Q -- that is,
         RHO(3) is the size of the last step.  The caller can decide to
         terminate the iterations, to pursue the iterations or to restart the
         algorithm.

      SPL_CG_AP: The caller must compute the matrix vector multiplication
          Q = A.P (i.e., store into vector Q the result of multiplying vector
          P by the operator A) and call spl_cgstp again.

      SPL_CG_PRECOND: The caller must compute and store the preconditioned
         residuals into Z (given the current residuals which are in R) and
         call spl_cgstp again.

      SPL_CG_FINISH: No further progress are possible either because the
         exact solution has been found or because rounding errors prevent
         further progress.

      SPL_CG_ERROR: An error has been detected by the algorithm: invalid
         parameter or non-positive definitiveness of the operator A or of the
         preconditioner.

      For consistency reasons, SPL_CG_NEWX is returned after internal
      initialization, i.e. X is just the initial solution x0 and R is set
      with r = b - A.x0.


      The routine spl_cgstp uses reverse communication and stores its local
      variables into its arguments.  As a consequence, P, Q, R, X, Z, RHO and
      STATE must be provided as variables (not expressions) and must not be
      modified between calls to spl_cgstp.  Another requirement is that P, Q,
      R, X, Z and RHO must all be of the same type (float or double).

      See spl_cgsv function for a higher level driver.


   EXAMPLE

    In the following piece of code, we assume that the variables A and C store
    respectively the coefficients of the matrix A and of the preconditioner.
    Note that this code is not meant to be efficient, it is just an example to
    illustrate a typical usage of spl_cgstp.

      x = ...; // initial set of parameters
      r = b;   // copy b into r
      p = q = z = array(structof(r), dimsof(r));
      n = numberof(r);
      rho = array(structof(r), 4);
      state = SPL_CG_RESTART;
      while (1) {
        spl_cgstp, n, p, q, r, x, z, rho, state;
        if (state == SPL_CG_AP) {
          q = a(+,)*p(+); // faster than a(,+)*p(+) when a is symmetric
        } else if (state == SPL_CG_PRECOND) {
          z = c(+,)*r(+); // faster than c(,+)*r(+) when c is symmetric
        } else if (state == SPL_CG_NEWX) {
          if (spl_nrm2(r) < 1e-3) {
            break; // convergence
          }
        } else {
          // unexpected (must be an error)
          error, swrite(format="error in spl_cgsv (%d)", state);
        }
      }


   SEE ALSO spl_cgsv.
*/

func spl_cgsv(a, b, x, precond=, maxiter=, verb=, view=, atol=, rtol=)
/* DOCUMENT x = spl_cgsv(a, b);
         or x = spl_cgsv(a, b, x0);

      Solve the linear system A.x = b using the conjugate gradient method.
      Argument A is the "left-hand-side matrix", it is a symmetric positive
      definite linear operator which can be provided as either a numerical
      array or as a function.  Argument B is the "right-hand-side vector", it
      is a floating point array.  If A is an array, mvmult(A, p) is used to
      compute A.p; otherwise, A is used as a function and A(p) should return
      the result of A.p (with p any "vector" of same type and dimensions as
      b).

      Optional argument X0 may be specified to provide an initial solution.
      By default, the initial solution is equal to zero.

      Keyword PRECOND may be used to specify a preconditioner; that is a
      linear operator which approximates the inverse of A (up to an arbitrary
      positive multiplicative factor).  Like A, PRECOND may be either an array
      or a function. By default, no preconditioner is used.

      Keyword MAXITER can be used to specify the maximum number of conjugate
      gradient iterations.  By default, MAXITER is equal to min(30, N) where
      N = numberof(B).

      Keyword VERB can be used to have the algorithm print out some
      information every VERB iterations.  In addition, keyword VIEW may be set
      with a subroutine to call every VERB iterations with the current
      solution X as unique argument.

      Keywords RTOL and ATOL may be used to specify the relative and absolute
      tolerances for the convergence.  The algorithm stops when the Euclidean
      norm of the residuals r = b - A.x becomes less than a level TOL.  If
      neither ATOL nor RTOL are specified, TOL = 1E-6*||R0|| where ||R0|| is
      the Euclidean norm of the initial residuals.  Otherwise, if only ATOL is
      specified, then TOL = ATOL.  Otherwise, if only RTOL is specified,
      TOL = RTOL*||R0||.  Otherwise, if ATOL and RTOL are both specified,
      TOL = max(ATOL,RTOL*||R0||).


   SEE ALSO: mvmult, spl_cgstp, spl_nrm2, LUsolve, SVsolve, QRsolve.
 */
{
  btype = structof(b);
  if ((type = btype) != double) {
    if (type == char || type == short || type == int || type == long) {
      type = float;
    }
    if (type != float) {
      error, "expecting non-complex numerical array for RHS vector B";
    }
  }
  lhs_is_array = is_array(a);
  if (lhs_is_array && (atype = structof(a)) != type) {
    if (atype == double) {
      type = double;
    } else if (atype != float && atype != char && atype != short &&
               atype != int && atype != long) {
      error, "expecting an operator or a non-complex numerical array for LHS matrix A";
    }
  }
  precond_is_array = is_array(precond);
  if (precond_is_array && (ptype = structof(precond)) != type) {
    if (ptype == double) {
      type = double;
    } else if (ptype == float || ptype == char || ptype == short ||
               ptype == int || ptype == long) {
      precond = type(precond);
    } else {
      error, "expecting an operator or a non-complex numerical array for preconditioner";
    }
  }
  if (lhs_is_array && atype != type) {
    a = type(a);
  }
  if (btype != type) {
    b = type(b);
  }
  n = numberof(b);
  dims = dimsof(b);
  if (is_void(x)) {
    state = SPL_CG_START;
    x = array(type, dims);
  } else {
    xdims = dimsof(x);
    if (numberof(xdims) != numberof(dims) || anyof(xdims != dims)) {
      error, "bad dimension list for X";
    }
    xtype = structof(x);
    if (xtype != type) {
      if (xtype == double) {
        write, format="%s\n", "WARNING - converting X to single precision";
      } else if (xtype != char && type != short && type != int &&
                 type != long && xtype != float) {
        error, "bad data type for X";
      }
      x = type(x);
    }
    state = (anyof(x) ? SPL_CG_RESTART : SPL_CG_START);
  }
  p = q = r = b; // FIXME: slight optimization: eq_nocopy, r, b;
  rho = array(type, 4);
  if (is_void(maxiter)) {
    maxiter = min(n, 30);
  }
  iter = -1;
  eval = 0;
  if (verb) {
    write, format="%s\n%s\n",
      " Iter    Residual Norm   CPU [s]      Alpha       Beta",
      "--------------------------------------------------------";
    format = "%5d %16.8e %9.3f  %10.2e  %10.2e\n";
    elapsed = array(double, 3);
    timer, elapsed;
    cpu_start = elapsed(1);
  }
  for (;;) {
    /* Call reverse commnunication version of conjugate gradient
       routine, also make sure STATE (which is passed by address) is
       of proper type. */
    state = int(state)
    if (precond) {
      spl_cgstp, n, p, q, r, x, q, rho, state;
    } else {
      spl_cgstp, n, p, q, r, x, rho, state;
    }

    /* Behave according to output value of STATE. */
    if (state == SPL_CG_AP) {
      q = (lhs_is_array ? mvmult(a, p) : type(a(p)));
      ++eval;
    } else if (state == SPL_CG_PRECOND) {
      q = (precond_is_array ? mvmult(precond, r) : type(precond(r)));
    } else if (state == SPL_CG_NEWX || state == SPL_CG_FINISH) {
      ++iter;
      rnrm2 = spl_nrm2(r); // FIXME: use RHO if no preconditioner
      if (iter == 0) {
        if (! is_void(rtol)) {
          if (is_void(atol)) {
            atol = rtol*rnrm2;
          } else {
            atol = max(atol, rtol*rnrm2);
          }
        } else if (is_void(atol)) {
          atol = 1e-6*rnrm2;
        }
      } else if (iter % n == 0) {
        r = b;
        state = SPL_CG_RESTART;
      }
      convergence = (rnrm2 <= atol);
      finish = (convergence || iter >= maxiter);
      prnt = (verb && (finish || ! (iter % verb)));
      if (prnt) {
        timer, elapsed;
        cpu = elapsed(1) - cpu_start;
        write, format=format, iter, rnrm2, cpu, rho(3), rho(4);
        if (view) {
          view, x;
        }
      }
      if (finish) {
        if (! convergence) {
          write, format="%s\n", "WARNING - too many iterations";
        }
        return x;
      }
    } else {
      error, "error in SPL_CGSTP";
    }
  }
}

/*---------------------------------------------------------------------------*/
/* TRUST REGION */

local spl_gqtpar_info;
extern spl_gqtpar;
/* DOCUMENT   x = spl_gqtpar(a, b, delta, par, f, iter, status);
         or msg = spl_gqtpar_info(status);

     Given an N by N symmetric matrix A, an N-vector B, and a positive
     number DELTA, this function determines a vector X which approximately
     minimizes the quadratic function:

         f(x) = (1/2)*x'.A.x + b'.x

     subject to the Euclidean norm constraint:

         norm(x) <= delta.

     This function computes an approximation X and a Lagrange multiplier
     PAR such that either PAR is zero and

          norm(x) <= (1 + rtol)*delta,

     or PAR is positive and

          abs(norm(x) - delta) <= rtol*delta.

     If XSOL is the exact solution to the problem, the approximation X
     satisfies:

          f(x) <= ((1 - rtol)^2)*f(xsol)


   ARGUMENTS:

     A is a real 2-D array of dimension N by N or a 1-D array of length
       N*(N+1)/2.  If A is 2-D, the full upper triangle of A must contain the
       full upper triangle of the symmetric matrix A.  If A is 1-D, it must
       contains the full upper triangle of the symmetric matrix A.

     B is a real array of dimension N which specifies the linear term in the
       quadratic approximation.

     DELTA is a positive real value equal to the bound on the Euclidean norm
       of X.

     PAR is a variable.  On entry, PAR is an initial estimate of the Lagrange
       multiplier for the constraint norm(x) <= delta.  If PAR is undefined on
       entry, PAR = 0.0 is tried first.  On exit, PAR contains the final
       estimate of the multiplier.

     F is an optional variable.  On entry, F need not be specified.  On exit,
       F is set to f(X) at the output X.

     STATUS is an optional variable.  On entry STATUS need not be specified.
       On exit STATUS is set as follows:

          status = 1  The function value f(x) has the relative
                    accuracy specified by rtol.

          status = 2  The function value f(x) has the absolute
                    accuracy specified by atol.

          status = 3  Rounding errors prevent further progress.
                    On exit x is the best available approximation.

          status = 4  Failure to converge after itmax iterations.
                    On exit x is the best available approximation.

        Function spl_gqtpar_info(STATUS) can be used to query the explanation
        corresponding to the output value of STATUS.


   KEYWORDS:

     RTOL is the relative accuracy desired in the solution. Convergence
       occurs if:

            f(x) <= ((1 - rtol)^2)*f(xsol)

       By default RTOL=0.05; otherwise the value for RTOL must be a
       positive real scalar.

     ATOL is the absolute accuracy desired in the solution. Convergence
       occurs when:

            norm(x) <= (1 + rtol)*delta

            max(-f(x), -f(xsol)) <= atol

       By default ATOL = 1E-8; otherwise the value for ATOL is a real scalar.

     ITMAX is the maximum number of iterations.  By default ITMAX = 7;
       otherwise the value for ITMAX is an integer scalar.


   RESULT:

     The result X is a real array of dimension N set to the final estimate of
     the solution.


   REFERENCES:

     [1] Moré, J.J. & Sorensen, D.C., "Computing A Trust Region Step",
         SIAM J. Sci. Stat. Comp., vol. 4, pp. 553-572, 1983.
 */
func spl_gqtpar_info(status)
{
  if (status == 1n) return "The function value f(x) has the relative accuracy specified by rtol.";
  if (status == 2n) return "The function value f(x) has the absolute accuracy specified by atol.";
  if (status == 3n) return "Rounding errors prevent further progress.  On exit x is the best available approximation.";
  if (status == 4n) return "Failure to converge after itmax iterations.  On exit x is the best available approximation.";
  return "Unknown status value. (BUG?)";
}

/*---------------------------------------------------------------------------*/
/* LINE SEARCH */

extern spl_lnsrch_new;
extern spl_lnsrch_reset;
/* DOCUMENT ls = spl_lnsrch_new();
         or spl_lnsrch_reset, ls;

      The function spl_lnsrch_new creates a new line search object.  The
      object is initialized with default values and is automatically destroyed
      when no longer referenced by Yorick.

      The function spl_lnsrch_reset must be used after convergence of the line
      search and prior to starting a new line search.  This avoids to create a
      new object for every successive line searches.
         
   SEE ALSO: spl_lnsrch_set_criterion.
 */

local SPL_LNSRCH_WOLFE, SPL_LNSRCH_ARMIJO;
extern spl_lnsrch_set_criterion;
extern spl_lnsrch_get_criterion;
extern spl_lnsrch_set_defaults;
extern spl_lnsrch_set_xtol;
extern spl_lnsrch_get_xtol;
extern spl_lnsrch_set_ftol;
extern spl_lnsrch_get_ftol;
extern spl_lnsrch_set_gtol;
extern spl_lnsrch_get_gtol;
/* DOCUMENT spl_lnsrch_set_criterion, ls, crit;
         or crit = spl_lnsrch_get_criterion(ls);
         or spl_lnsrch_set_ftol, ls, ftol;
         or ftol = spl_lnsrch_get_ftol(ls);
         or spl_lnsrch_set_gtol, ls, gtol;
         or gtol = spl_lnsrch_get_gtol(ls);
         or spl_lnsrch_set_xtol, ls, xtol;
         or xtol = spl_lnsrch_get_xtol(ls);
         or spl_lnsrch_set_defaults, ls;

     These functions set or get the value of the parameters used to define the
     stopping criteria for line search with object LS.

     CRIT defines the conditions for a successfull step, it can be
     SPL_LNSRCH_WOLFE to apply the two Wolfe's conditions (sufficient function
     decrease and curvature condition) or SPL_LNSRCH_ARMIJO to only apply
     Armijo's condition of sufficient function decrease.
     
     FTOL is the parameter for the sufficient decrease condition (first
     Wolfe's condition or Armijo's condition) which requires that:

         f(x + stp*d) <= f(x) + ftol*stp*<d,g(x)>

     where f(x) is the objective function, g(x) is its gradient, x are the
     variables at the start of the line search, d is the search direction, stp
     is the setp lenght and <a,b> denotes scalar product of a and b.

     GTOL is the parameter for the curvature condition (second Wolfe's
     condition) which requires that:
     
         abs(<d,g(x + stp*d)>) <= gtol*abs(<d,g(x)>)

     XTOL is the minimum relative size of the uncertainty interval for the
     step size.

     The subroutine spl_lnsrch_set_defaults sets default values for FTOL, GTOL
     and XTOL.
     
   SEE ALSO: spl_lnsrch_new.
 */
SPL_LNSRCH_WOLFE  = 0n;
SPL_LNSRCH_ARMIJO = 1n;

local SPL_LNSRCH_ERROR, SPL_LNSRCH_START, SPL_LNSRCH_FG, SPL_LNSRCH_CONV, SPL_LNSRCH_WARN, SPL_LNSRCH_NONE, SPL_LNSRCH_ERROR_ADDRESS, SPL_LNSRCH_ERROR_CORRUPTED, SPL_LNSRCH_ERROR_STPMIN_LT_ZERO, SPL_LNSRCH_ERROR_STPMAX_LT_STPMIN, SPL_LNSRCH_ERROR_NOT_DESCENT, SPL_LNSRCH_ERROR_STP_GT_STPMAX, SPL_LNSRCH_ERROR_STP_LT_STPMIN, SPL_LNSRCH_ERROR_STP_OUTSIDE_RANGE, SPL_LNSRCH_CONVERGENCE_ARMIJO, SPL_LNSRCH_CONVERGENCE_STRONG_WOLFE, SPL_LNSRCH_WARNING_STP_EQ_STPMIN, SPL_LNSRCH_WARNING_STP_EQ_STPMAX, SPL_LNSRCH_WARNING_XTOL_TEST_SATISFIED, SPL_LNSRCH_WARNING_ROUNDING_ERRORS;
extern spl_lnsrch_get_task;
extern spl_lnsrch_get_info_code;
extern spl_lnsrch_get_info_text;
/* DOCUMENT spl_lnsrch_get_task(ls);
         or spl_lnsrch_get_info_code(ls);
         or spl_lnsrch_get_info_text(ls);

     The function spl_lnsrch_get_task returns the current task in line search
     object LS.  The possible values are:

       SPL_LNSRCH_START - Line search in progress or not yet started.
       SPL_LNSRCH_FG    - The caller is requested to compute the function
                          value and the gradient at the new point and then
                          call the (FIXME:) spl_lnsrch subroutine again.
       SPL_LNSRCH_CONV  - Convergence criteria are satisfied.
       SPL_LNSRCH_WARN  - The solution cannot be improved for some reason.
       SPL_LNSRCH_ERROR - Some major error occured.

     When the tast is SPL_LNSRCH_CONV, SPL_LNSRCH_WARN, or SPL_LNSRCH_ERROR,
     the functions spl_lnsrch_get_info_code and spl_lnsrch_get_info_text can
     be used to obtain more information.  The former returns an integer code
     while the latter returns a textual information.  The following codes are
     implemented:

       SPL_LNSRCH_NONE - Line search not yet started.
       SPL_LNSRCH_ERROR_ADDRESS - Invalid address.
       SPL_LNSRCH_ERROR_CORRUPTED - Corrupted structure.
       SPL_LNSRCH_ERROR_STPMIN_LT_ZERO - Minimum step length is less than zero.
       SPL_LNSRCH_ERROR_STPMAX_LT_STPMIN - Maximum step length is less than
               minimum one.
       SPL_LNSRCH_ERROR_DERIV_GE_ZERO - Search direction is not a descent.
       SPL_LNSRCH_ERROR_STP_GT_STPMAX - Step length is above superior bound.
       SPL_LNSRCH_ERROR_STP_LT_STPMIN - Step length is below inferior bound.
       SPL_LNSRCH_ERROR_STP_OUTSIDE_RANGE - Step length is outside the bracketed
               interval.
       SPL_LNSRCH_CONVERGENCE_ARMIJO - Line search converged according to
               Armijo criterion.
       SPL_LNSRCH_CONVERGENCE_STRONG_WOLFE - Line search converged according to
               strong Wolfe criteria.
       SPL_LNSRCH_WARNING_STP_EQ_STPMIN - Step length at inferior bound.
       SPL_LNSRCH_WARNING_STP_EQ_STPMAX - Step length at superior bound.
       SPL_LNSRCH_WARNING_XTOL_TEST_SATISFIED - Relative uncertainty of step
               length is within tolerances.
       SPL_LNSRCH_WARNING_ROUNDING_ERRORS - Rounding errors prevent progress.

     
   SEE ALSO: spl_lnsrch_new.
 */

/* Task codes. */
SPL_LNSRCH_ERROR = -1n;
SPL_LNSRCH_START =  0n;
SPL_LNSRCH_FG    =  1n;
SPL_LNSRCH_CONV  =  2n;
SPL_LNSRCH_WARN  =  3n;

/* Information codes. */
SPL_LNSRCH_NONE = 0n;
SPL_LNSRCH_ERROR_ADDRESS = 1n;
SPL_LNSRCH_ERROR_CORRUPTED = 2n;
SPL_LNSRCH_ERROR_STPMIN_LT_ZERO = 3n;
SPL_LNSRCH_ERROR_STPMAX_LT_STPMIN = 4n;
SPL_LNSRCH_ERROR_NOT_DESCENT = 5n;
SPL_LNSRCH_ERROR_STP_GT_STPMAX = 6n;
SPL_LNSRCH_ERROR_STP_LT_STPMIN = 7n;
SPL_LNSRCH_ERROR_STP_OUTSIDE_RANGE = 8n;
SPL_LNSRCH_CONVERGENCE_ARMIJO = 9n;
SPL_LNSRCH_CONVERGENCE_STRONG_WOLFE = 10n;
SPL_LNSRCH_WARNING_STP_EQ_STPMIN = 11n;
SPL_LNSRCH_WARNING_STP_EQ_STPMAX = 12n;
SPL_LNSRCH_WARNING_XTOL_TEST_SATISFIED = 13n;
SPL_LNSRCH_WARNING_ROUNDING_ERRORS = 14n;

extern spl_lnsrch_get_ginit;
extern spl_lnsrch_get_finit;
extern spl_lnsrch_get_stx;
extern spl_lnsrch_get_fx;
extern spl_lnsrch_get_gx;
extern spl_lnsrch_get_sty;
extern spl_lnsrch_get_fy;
extern spl_lnsrch_get_gy;
/* DOCUMENT 
     
   SEE ALSO:
 */


/*---------------------------------------------------------------------------*/
/* INTERPOLATION AND B-SPLINES */

extern spl_interp_coefs_get_m;
extern spl_interp_coefs_get_n;
extern spl_interp_coefs_get_p;
extern spl_interp_coefs_get_j;
extern spl_interp_coefs_get_w;
/* DOCUMENT spl_interp_coefs_get_m(ws);
         or spl_interp_coefs_get_n(ws);
         or spl_interp_coefs_get_p(ws);
         or spl_interp_coefs_get_j(ws);
         or spl_interp_coefs_get_w(ws);

     These functions query the contents of the set of interpolation
     coefficients WS.  M is the number of interpolation coordinates, N is the
     size of the interpolation grid, P is the number of grid nodes needed for
     a single interpolated value, J is the array of 0-based grid indices of
     the non-zero interpolation coefficients and W is the array of non-zero
     interpolation coefficients.


   SEE ALSO
     spl_interp_coefs_new_spline.
 */

extern spl_interp_coefs_scale;
/* DOCUMENT spl_interp_coefs_scale(ws, a);
         or spl_interp_coefs_scale, ws, a;

     This function rescale the coefficients of the set of interpolation
     coefficients WS by a factor A.  If called as a subroutine, the operation
     is done in-place.


   SEE ALSO
     spl_interp_coefs_new_spline.
 */

extern spl_interp_coefs_new_spline;
/* DOCUMENT ws = spl_interp_coefs_new_spline(x, n, deg);

     This function creates a new set of interpolation coefficients for
     interpolation by B-splines of degree DEG.  X is the array of
     interpolation coordinates and N is the lenght of the spline grid.  The
     coordinates X are in grid units: 0 is the first grid node, N-1 is the
     last one.  Mirror boundary conditions are applied for coordinates outside
     the grid size.

     The returned object can be used as a function to apply the corresponding
     linear operator (or its transpose if TR is true):

        ws(x)       is the same as     spl_interp_apply(x, ws)
        ws(x, tr)   is the same as     spl_interp_apply(x, ws, tr)

     The members of the returned object can be queried using the "dot"
     dereference operator, for instance:

        ws.m        is the same as     spl_interp_coefs_get_m(ws)
        ws.j        is the same as     spl_interp_coefs_get_j(ws)


   SEE ALSO
     spl_interp_apply, spl_interp_coefs_get_m.
 */

extern spl_interp_apply;
/* DOCUMENT a = spl_interp_apply(c, ws);
         or a = spl_interp_apply(c, ws, transp);
         or a = spl_interp_apply(c, ws1, ..., wsN);
         or a = spl_interp_apply(c, ws1, ..., wsN, transp);

     Apply interpolation operator on array C.  The WS's are sets of
     interpolation coefficients (e.g. as returned by
     spl_interp_coefs_new_spline) along each dimension of C.  The two first
     examples are the same as: a = ws(c) and a = ws(c, transp).

     Optional last argument TRANSP can be set true to apply the transpose
     linear operator.

     The first set of coefficients (WS1) can be nil to not interpolate along
     the initial dimension of C.


   SEE ALSO
     spl_interp_coefs_new_spline, spl_spline_resampler.
 */

extern spl_spline_samples_to_coefficients;
/* DOCUMENT c = spl_spline_samples_to_coefficients(a, deg, ...);
         or spl_spline_samples_to_coefficients, a, deg, ...;

     Convert sampled array A into B-spline coefficients. DEG and subsequent
     arguments give the degrees of the B-splines.  A single scalar DEG, set
     the degree for all the dimensions of A; otherwise, DEG concatenated with
     all subsequent arguments set the degree of the splines for the first
     dimensions of A.  Note that B-splines of degree 0 or 1 are also
     interpolation functions (equal to zero for non-zero integer offsets),
     hence specifying a degree of 0 or 1 along a particular dimension has no
     effect.

     When called as a subroutine, the operation is done in place.


   SEE ALSO
     spl_interp_coefs_new_spline, spl_spline_coefficients_to_samples.
 */

local _spl_interp_evaluator_1, _spl_interp_evaluator_2;
local _spl_interp_evaluator_3, _spl_interp_evaluator_4;
func spl_spline_interpolator(.., scale=)
/* DOCUMENT op = spl_spline_interpolator(x, n, deg);
         or op = spl_spline_interpolator(x1, n1, deg1, x2, n2, ge2, ...);

      Build a linear operator to perform interpolation by B-splines.
      Interpolation along i-th dimension is specified by a triplet (Xi, Ni,
      DEGi) with Xi the interpolation coordinates, Ni the number of
      interpolation nodes along that dimension and DEGi the degree of the
      B-spline. The interpolation coordinates are in units of the grid node
      which span the range [0, Ni-1], mirror boundary conditions are used for
      coordinates outside this range.

      The returned object can be used a linear operator:

         OP(c)     to apply the linear operator; that is, interpolation of
                   N1-by-N2-by-... array C of spline coefficients;

         OP(a, 1)  to apply the transpose linear operator to array A of
                   dimensions: numberof(X1)-by-numberof(X2)-by-...;

       Keyword SCALE can be used to multiply the interpolation by this factor;
       by default SCALE=1.


   SEE ALSO: spl_interp_coefs_new_spline, spl_interp_coefs_scale.
 */
{
  if (is_void(scale)) scale = 1.0;
  obj = h_new();
  i = 0;
  fmt = "ws%d";
  while (more_args()) {
    if (++i > 4) error, "too many dimensions";
    ws = spl_interp_coefs_new_spline(next_arg(), next_arg(), next_arg());
    if (scale != 1.0) spl_interp_coefs_scale, ws, scale;
    h_set, obj, swrite(format=fmt, i), ws;
  }
  if (i < 1) error, "not enough arguments";
  h_evaluator, obj, swrite(format="_spl_interp_evaluator_%d", i);
  return obj;
}
func _spl_interp_evaluator_1(self, arg, job)
{
  return spl_interp_apply(arg, self.ws1, (job ? 1 : 0));
}
func _spl_interp_evaluator_2(self, arg, job)
{
  return spl_interp_apply(arg, self.ws1, self.ws2, (job ? 1 : 0));
}
func _spl_interp_evaluator_3(self, arg, job)
{
  return spl_interp_apply(arg, self.ws1, self.ws2, self.ws3, (job ? 1 : 0));
}
func _spl_interp_evaluator_4(self, arg, job)
{
  return spl_interp_apply(arg, self.ws1, self.ws2, self.ws3, self.ws4,
                          (job ? 1 : 0));
}

/*
 * Local Variables:
 * mode: Yorick
 * tab-width: 8
 * indent-tabs-mode: nil
 * c-basic-offset: 2
 * fill-column: 78
 * coding: utf-8
 * End:
 */
